@@ -4,11 +4,35 @@ use strict;
 use warnings;
 use 5.010;
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
+
+use Exporter 'import';
+our @EXPORT_OK = qw(
+    INLINE_NONE INLINE_GETTER INLINE_SETTER 
+    INLINE_HV_GETTER INLINE_HV_SETTER
+);
+our %EXPORT_TAGS = (
+    inline => [qw(INLINE_NONE INLINE_GETTER INLINE_SETTER 
+                  INLINE_HV_GETTER INLINE_HV_SETTER)],
+);
 
 require XSLoader;
 XSLoader::load('XS::JIT::Builder', $VERSION);
 
+sub write_file {
+    my ($self, $filename) = @_;
+
+    die "write_file requires a filename" unless defined $filename;
+
+    open my $fh, '>', $filename
+        or die "Cannot open '$filename' for writing: $!";
+    print $fh $self->code;
+    close $fh
+        or die "Cannot close '$filename': $!";
+
+    return $self;
+}
+
 =head1 NAME
 
 XS::JIT::Builder - interface for building XS/C code strings
@@ -96,6 +120,25 @@ Clear the builder and reset state for reuse.
     $b->reset;
     # Builder is now empty, can build new code
 
+=item write_file($filename)
+
+Write the accumulated C code to a file. This is useful for:
+
+=over 4
+
+=item * Distribution - Generate the XS file once, ship it with your module
+
+=item * Debugging - Inspect the generated code
+
+=item * Performance - Skip runtime JIT compilation entirely
+
+=back
+
+    $b->write_file('MyModule.c');
+    # Now you can compile MyModule.c as a regular XS/C extension
+
+Returns C<$self> to enable method chaining.
+
 =back
 
 =head2 Low-level Output
@@ -700,6 +743,284 @@ or simple attribute name strings.
     package Dog;
     use parent -norequire => 'Animal';
 
+=head2 Ultra-fast Array-based Objects with Inline Ops
+
+For maximum performance, use array-based objects with inline ops.
+This bypasses XS call overhead entirely at compile time:
+
+    use XS::JIT;
+    use XS::JIT::Builder qw(:inline);
+    
+    # Generate op-based accessors
+    my $b = XS::JIT::Builder->new;
+    $b->op_ro_accessor('get_name', 0);  # slot 0 is read-only
+    $b->op_rw_accessor('age', 1);       # slot 1 is read-write
+    
+    XS::JIT->compile(
+        code      => $b->code,
+        name      => 'Cat',
+        cache_dir => $cache,
+        functions => {
+            'Cat::name' => { source => 'get_name', is_xs_native => 1 },
+            'Cat::age'  => { source => 'age', is_xs_native => 1 },
+        },
+    );
+    
+    # Register inline ops for compile-time optimization
+    XS::JIT::Builder::inline_init();
+    XS::JIT::Builder::inline_register(\&Cat::name, INLINE_GETTER, 0);
+    XS::JIT::Builder::inline_register(\&Cat::age, INLINE_SETTER, 1);
+    
+    # Now function calls are replaced with custom ops at compile time!
+    package Cat;
+    sub new { bless [$_[1], $_[2]], $_[0] }
+    
+    package main;
+    my $cat = Cat->new('Whiskers', 3);
+    say $cat->name;      # Inline op - no XS call overhead
+    $cat->age(4);        # Inline setter
+
+=head2 Inline Op Types
+
+The following constants are available via C<:inline> export tag:
+
+=over 4
+
+=item INLINE_NONE (0)
+
+No inlining.
+
+=item INLINE_GETTER (1)
+
+Read-only slot accessor. Replaces C<< $obj->name >> with a custom op
+that reads directly from C<< $obj->[slot] >>.
+
+=item INLINE_SETTER (2)
+
+Read-write slot accessor. Supports both getter and setter modes.
+
+=item INLINE_HV_GETTER (3)
+
+Read-only hash accessor (not yet implemented).
+
+=item INLINE_HV_SETTER (4)
+
+Read-write hash accessor (not yet implemented).
+
+=back
+
+=head2 Custom Op Builder Methods
+
+These methods generate code for building custom Perl ops, enabling
+compile-time optimization through call checkers.
+
+=head3 PP Function Builders
+
+PP (push-pop) functions are the runtime handlers for custom ops.
+
+=over 4
+
+=item pp_start($name)
+
+Start a pp function definition.
+
+    $b->pp_start('pp_my_op');
+    # Generates: OP* pp_my_op(pTHX) {
+
+=item pp_end()
+
+End a pp function with C<return NORMAL>.
+
+    $b->pp_end;
+    # Generates: return NORMAL; }
+
+=item pp_dsp()
+
+Add dSP declaration for stack access.
+
+    $b->pp_dsp;
+    # Generates: dSP;
+
+=item pp_get_self()
+
+Get the invocant from the stack without popping.
+
+    $b->pp_get_self;
+    # Generates: SV* self = TOPs;
+
+=item pp_pop_self()
+
+Pop the invocant from the stack.
+
+    $b->pp_pop_self;
+    # Generates: SV* self = POPs;
+
+=item pp_pop_sv($name)
+
+Pop an SV from the stack.
+
+    $b->pp_pop_sv('value');
+    # Generates: SV* value = POPs;
+
+=item pp_pop_nv($name)
+
+Pop a numeric value from the stack.
+
+    $b->pp_pop_nv('amount');
+    # Generates: NV amount = POPn;
+
+=item pp_pop_iv($name)
+
+Pop an integer value from the stack.
+
+    $b->pp_pop_iv('count');
+    # Generates: IV count = POPi;
+
+=item pp_get_slots()
+
+Get the slots array from self (for array-based objects).
+
+    $b->pp_get_slots;
+    # Generates: AV* slots = (AV*)SvRV(self);
+
+=item pp_slot($name, $index)
+
+Access a specific slot from the slots array.
+
+    $b->pp_slot('name_sv', 0);
+    # Generates: SV* name_sv = *av_fetch(slots, 0, 0);
+
+=item pp_return_sv($expr)
+
+Return an SV to the stack.
+
+    $b->pp_return_sv('name_sv');
+    # Generates: SETs(name_sv); RETURN;
+
+=item pp_return_nv($expr)
+
+Return a numeric value.
+
+    $b->pp_return_nv('result');
+    # Generates: SETn(result); RETURN;
+
+=item pp_return_iv($expr)
+
+Return an integer value.
+
+    $b->pp_return_iv('count');
+    # Generates: SETi(count); RETURN;
+
+=item pp_return_pv($expr)
+
+Return a string value.
+
+    $b->pp_return_pv('str');
+    # Generates: SETs(sv_2mortal(newSVpv(str, 0))); RETURN;
+
+=item pp_return()
+
+Return without modifying the stack.
+
+    $b->pp_return;
+    # Generates: return NORMAL;
+
+=back
+
+=head3 Call Checker Builders
+
+Call checkers run at compile time to replace subroutine calls with custom ops.
+
+=over 4
+
+=item ck_start($name)
+
+Start a call checker function.
+
+    $b->ck_start('ck_my_method');
+    # Generates: OP* ck_my_method(pTHX_ OP *entersubop, GV *namegv, SV *ckobj) {
+
+=item ck_end()
+
+End a call checker with fallback.
+
+    $b->ck_end;
+    # Generates: return ck_entersub_args_proto_or_list(...); }
+
+=item ck_preamble()
+
+Add standard call checker preamble to extract arguments.
+
+    $b->ck_preamble;
+    # Extracts pushmark, args, method from the optree
+
+=item ck_build_unop($pp_func, $targ_expr)
+
+Build a custom unary op (one argument).
+
+    $b->ck_build_unop('pp_my_getter', 'slot_num');
+    # Generates code to create OP_CUSTOM with the specified pp function
+
+=item ck_build_binop($pp_func, $targ_expr)
+
+Build a custom binary op (two arguments).
+
+    $b->ck_build_binop('pp_my_setter', 'slot_num');
+
+=item ck_fallback()
+
+Fall through to default call checking.
+
+    $b->ck_fallback;
+    # Generates: return ck_entersub_args_proto_or_list(...);
+
+=back
+
+=head3 XOP Helpers
+
+XOP (extended op) declarations register custom ops with Perl.
+
+=over 4
+
+=item xop_declare($name, $pp_func, $desc)
+
+Declare a custom op descriptor.
+
+    $b->xop_declare('my_xop', 'pp_my_op', 'my custom op');
+    # Generates static XOP declaration and registration
+
+=item register_checker($cv_expr, $ck_func, $ckobj_expr)
+
+Register a call checker for a CV.
+
+    $b->register_checker('cv', 'ck_my_method', 'newSViv(slot)');
+    # Generates: cv_set_call_checker(cv, ck_my_method, ckobj);
+
+=back
+
+=head3 Custom Op Example
+
+    my $b = XS::JIT::Builder->new;
+    
+    # Declare the XOP
+    $b->xop_declare('slot_getter_xop', 'pp_slot_getter', 'slot getter');
+    
+    # Build the pp function
+    $b->pp_start('pp_slot_getter')
+      ->pp_dsp
+      ->pp_get_self
+      ->pp_get_slots
+      ->line('IV slot = PL_op->op_targ;')
+      ->pp_slot('val', 'slot')
+      ->pp_return_sv('val')
+      ->pp_end;
+    
+    # Build the call checker
+    $b->ck_start('ck_slot_getter')
+      ->ck_preamble
+      ->ck_build_unop('pp_slot_getter', 'SvIV(ckobj)')
+      ->ck_end;
+
 =head1 SEE ALSO
 
 L<XS::JIT> - The main JIT compiler module
