@@ -38,6 +38,18 @@ typedef struct {
     STRLEN len;
 } XS_JIT_Attr;
 
+/* ============================================
+ * Inline op types for cv_set_call_checker
+ * ============================================ */
+
+typedef enum {
+    XS_JIT_INLINE_NONE      = 0,
+    XS_JIT_INLINE_GETTER    = 1,  /* Read-only slot accessor */
+    XS_JIT_INLINE_SETTER    = 2,  /* Read-write slot accessor */
+    XS_JIT_INLINE_HV_GETTER = 3,  /* Hash-based read-only accessor */
+    XS_JIT_INLINE_HV_SETTER = 4   /* Hash-based read-write accessor */
+} XS_JIT_InlineType;
+
 /* ============================================
  * Lifecycle
  * ============================================ */
@@ -343,6 +355,127 @@ void xs_jit_rw_accessor(pTHX_ XS_JIT_Builder* b, const char* func_name, const ch
 /* Generate a constructor (attrs is NULL-terminated array) */
 void xs_jit_constructor(pTHX_ XS_JIT_Builder* b, const char* func_name, const char* class_name, XS_JIT_Attr* attrs, int num_attrs);
 
+/* ============================================
+ * Op-based accessors (for array-based objects with inline ops)
+ * These are faster than hash-based accessors and can be
+ * combined with inline ops for compile-time optimization
+ * ============================================ */
+
+/* Generate a read-only op accessor: $obj->[slot] */
+void xs_jit_op_ro_accessor(pTHX_ XS_JIT_Builder* b, const char* func_name, IV slot);
+
+/* Generate a read-write op accessor: $obj->[slot] with optional setter */
+void xs_jit_op_rw_accessor(pTHX_ XS_JIT_Builder* b, const char* func_name, IV slot);
+
+/* ============================================
+ * Custom op builder (for defining arbitrary inline ops)
+ * 
+ * Usage:
+ *   xs_jit_pp_start(b, "my_op");
+ *   xs_jit_pp_get_self(b);
+ *   xs_jit_line(b, "NV result = ...");
+ *   xs_jit_pp_return_nv(b, "result");
+ *   xs_jit_pp_end(b);
+ * ============================================ */
+
+/* Start a pp (push-pop) function - the runtime part of a custom op */
+void xs_jit_pp_start(pTHX_ XS_JIT_Builder* b, const char* name);
+
+/* End a pp function */
+void xs_jit_pp_end(pTHX_ XS_JIT_Builder* b);
+
+/* pp preamble: dSP; */
+void xs_jit_pp_dsp(pTHX_ XS_JIT_Builder* b);
+
+/* Get self from top of stack: SV* self = TOPs; */
+void xs_jit_pp_get_self(pTHX_ XS_JIT_Builder* b);
+
+/* Pop self from stack: SV* self = POPs; */
+void xs_jit_pp_pop_self(pTHX_ XS_JIT_Builder* b);
+
+/* Pop a value: SV* name = POPs; */
+void xs_jit_pp_pop_sv(pTHX_ XS_JIT_Builder* b, const char* name);
+
+/* Pop as NV: NV name = POPn; */
+void xs_jit_pp_pop_nv(pTHX_ XS_JIT_Builder* b, const char* name);
+
+/* Pop as IV: IV name = POPi; */
+void xs_jit_pp_pop_iv(pTHX_ XS_JIT_Builder* b, const char* name);
+
+/* Get array slots from self: SV** ary = AvARRAY((AV*)SvRV(self)); */
+void xs_jit_pp_get_slots(pTHX_ XS_JIT_Builder* b);
+
+/* Get slot value: SV* name = ary[slot] ? ary[slot] : &PL_sv_undef; */
+void xs_jit_pp_slot(pTHX_ XS_JIT_Builder* b, const char* name, IV slot);
+
+/* Set result and return: SETs(sv); return NORMAL; */
+void xs_jit_pp_return_sv(pTHX_ XS_JIT_Builder* b, const char* sv_expr);
+
+/* Return mortal NV: SETs(sv_2mortal(newSVnv(expr))); return NORMAL; */
+void xs_jit_pp_return_nv(pTHX_ XS_JIT_Builder* b, const char* nv_expr);
+
+/* Return mortal IV: SETs(sv_2mortal(newSViv(expr))); return NORMAL; */
+void xs_jit_pp_return_iv(pTHX_ XS_JIT_Builder* b, const char* iv_expr);
+
+/* Return mortal PV: SETs(sv_2mortal(newSVpv(expr, 0))); return NORMAL; */
+void xs_jit_pp_return_pv(pTHX_ XS_JIT_Builder* b, const char* pv_expr);
+
+/* Just return NORMAL; */
+void xs_jit_pp_return(pTHX_ XS_JIT_Builder* b);
+
+/* ============================================
+ * Call checker builder (compile-time rewriting)
+ * ============================================ */
+
+/* Start a call checker function */
+void xs_jit_ck_start(pTHX_ XS_JIT_Builder* b, const char* name);
+
+/* End a call checker function */
+void xs_jit_ck_end(pTHX_ XS_JIT_Builder* b);
+
+/* Standard call checker preamble - extract args from entersub */
+void xs_jit_ck_preamble(pTHX_ XS_JIT_Builder* b);
+
+/* Build and return a UNOP custom op */
+void xs_jit_ck_build_unop(pTHX_ XS_JIT_Builder* b, const char* pp_func, const char* targ_expr);
+
+/* Build and return a BINOP custom op */
+void xs_jit_ck_build_binop(pTHX_ XS_JIT_Builder* b, const char* pp_func, const char* targ_expr);
+
+/* Return the original entersubop (fallback, no optimization) */
+void xs_jit_ck_fallback(pTHX_ XS_JIT_Builder* b);
+
+/* ============================================
+ * XOP registration helpers
+ * ============================================ */
+
+/* Generate XOP struct and registration code */
+void xs_jit_xop_declare(pTHX_ XS_JIT_Builder* b, const char* name, const char* pp_func, const char* desc);
+
+/* Generate cv_set_call_checker_flags call */
+void xs_jit_register_checker(pTHX_ XS_JIT_Builder* b, const char* cv_expr, const char* ck_func, const char* ckobj_expr);
+
+/* ============================================
+ * Inline op support (compile-time optimization)
+ * 
+ * Inline ops replace function calls with custom ops at compile time,
+ * bypassing XS call overhead entirely. This provides ~2x speedup.
+ * 
+ * Usage:
+ *   CV* cv = get_cv("MyClass::name", 0);
+ *   xs_jit_inline_register(aTHX_ cv, XS_JIT_INLINE_GETTER, 0, NULL, 0);
+ * ============================================ */
+
+/* Initialize inline op subsystem (safe to call multiple times) */
+void xs_jit_inline_init(pTHX);
+
+/* Register an inline op for a CV */
+int xs_jit_inline_register(pTHX_ CV* cv, XS_JIT_InlineType type, 
+                           IV slot, const char* key, STRLEN key_len);
+
+/* Check if a CV has an inline op registered */
+XS_JIT_InlineType xs_jit_inline_get_type(pTHX_ CV* cv);
+
 /* ============================================
  * Convenience macros
  * Assumes 'b' is the builder variable name
@@ -419,4 +552,7 @@ void xs_jit_constructor(pTHX_ XS_JIT_Builder* b, const char* func_name, const ch
 #define JIT_RO_ACCESSOR(f, a, l)    xs_jit_ro_accessor(aTHX_ b, f, a, l)
 #define JIT_RW_ACCESSOR(f, a, l)    xs_jit_rw_accessor(aTHX_ b, f, a, l)
 
+#define JIT_OP_RO(f, s)             xs_jit_op_ro_accessor(aTHX_ b, f, s)
+#define JIT_OP_RW(f, s)             xs_jit_op_rw_accessor(aTHX_ b, f, s)
+
 #endif /* XS_JIT_BUILDER_H */
